#ifndef FrameCPP_VERSION_6_FrTOC_HH
#define FrameCPP_VERSION_6_FrTOC_HH

#include <map>
#include <vector>

#include "framecpp/Common/FrameSpec.hh"
#include "framecpp/Common/FrTOC.hh"

#include "framecpp/Version6/FrameSpec.hh"
#include "framecpp/Version6/FrTOCAdcData.hh"
#include "framecpp/Version6/FrTOCData.hh"
#include "framecpp/Version6/FrTOCEvent.hh"
#include "framecpp/Version6/FrTOCProcData.hh"
#include "framecpp/Version6/FrTOCSerData.hh"
#include "framecpp/Version6/FrTOCSimData.hh"
#include "framecpp/Version6/FrTOCSimEvent.hh"
#include "framecpp/Version6/FrTOCStatData.hh"
#include "framecpp/Version6/FrTOCSummary.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {

    //===================================================================
    //===================================================================
    class FrTOC
      : public Common::FrTOC,
        public FrTOCData,
	public FrTOCStatData,
	public FrTOCAdcData,
	public FrTOCProcData,
	public FrTOCSimData,
	public FrTOCSerData,
	public FrTOCSummary,
	public FrTOCEvent,
	public FrTOCSimEvent
    {
    public:
      typedef INT_8U				position_type;
      typedef std::vector< position_type >	position_list_type;

      class StatTypeKey {
      public:
	std::string	s_name;
	std::string	s_detector;

	StatTypeKey( const std::string& Name,
		     const std::string& Detector );
	bool operator<( const StatTypeKey& RHS ) const;
	bool operator==( const StatTypeKey& RHS ) const;
	bool operator!=( const StatTypeKey& RHS ) const;
      };

      struct StatType_type
      {
	std::vector< INT_4U >	tStart;
	std::vector< INT_4U >	tEnd;
	std::vector< INT_4U >	version;
	position_list_type	positionStat;
      };

      class StatTypeKeyHash
      {
      public:
	size_t operator()( const StatTypeKey& Key ) const
	{
	  General::hash< const char* > h;

	  std::string key( Key.s_name );
	  key += Key.s_detector;
	  return h( key.c_str( ) );
	}
      };

      typedef General::hash_map< StatTypeKey, StatType_type, StatTypeKeyHash >
      MapStatType_type;

      static const INT_4U	NO_DATA_AVAILABLE;

      FrTOC( );

      static const char* const StructName( );

      static const Common::Description* StructDescription( );

      typedef Common::FrameSpec::Object object_type;

      virtual void IndexObject( const object_type* Object,
				std::streampos Position );

      virtual Common::FrameSpec::size_type
      Bytes( const Common::StreamBase& Stream ) const;

      virtual FrTOC* Create( ) const;

      virtual FrTOC* Create( Common::IStream& Stream ) const;

      virtual void FrStatDataQuery( const std::string& NamePattern,
				    const General::GPSTime& StartTime,
				    const General::GPSTime& EndTime,
				    const INT_4U Version,
				    Common::FrStatData::Query& Result ) const;

      virtual const char* const ObjectStructName( ) const;

      virtual void Write( Common::OStream& Stream ) const;
 
      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      const MapStatType_type&		GetStatType( ) const;

    protected:
      /// \brief Down grade an object
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      /// \brief Upgrade an object
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

    protected:
      //-----------------------------------------------------------------
      // Interface routines
      //-----------------------------------------------------------------
      virtual INT_4U nFrame( ) const;

      virtual const cmn_dt_container_type& dt( ) const;
      virtual const cmn_GTimeS_container_type& GTimeS( ) const;
      virtual const cmn_GTimeN_container_type& GTimeN( ) const;

      virtual INT_8U positionH( INT_4U FrameIndex ) const;

      virtual const cmn_name_container_type& nameADC( ) const;

      virtual INT_8U positionADC( INT_4U FrameIndex,
				  const std::string& Channel ) const;
      virtual INT_8U positionADC( INT_4U FrameIndex,
				  INT_4U Channel ) const;

      virtual const cmn_name_container_type& nameProc( ) const;

      virtual INT_8U positionProc( INT_4U FrameIndex,
				   const std::string& Channel ) const;
      virtual INT_8U positionProc( INT_4U FrameIndex,
				   INT_4U Channel ) const;

      virtual const cmn_name_container_type& nameSer( ) const;

      virtual INT_8U positionSer( INT_4U FrameIndex,
				  const std::string& Channel ) const;

      virtual const cmn_name_container_type& nameSim( ) const;

      virtual INT_8U positionSim( INT_4U FrameIndex,
				  const std::string& Channel ) const;

      virtual INT_4U nSH( ) const;
      virtual INT_2U SHid( INT_4U Offset ) const;
      virtual const std::string& SHname( INT_4U Offset ) const;

    private:
      using Common::FrameSpec::Object::Create;

      MapStatType_type		m_StatType;

      FrTOC( Common::IStream& Stream );

      template< typename ChannelType >
      INT_8U position_adc( INT_4U FrameIndex, ChannelType Channel ) const;

      template< typename ChannelType >
      INT_8U position_proc( INT_4U FrameIndex, ChannelType Channel ) const;

   };

    inline const char* const FrTOC::
    StructName( )
    {
      static const CHAR* class_name( "FrTOC" );
      return class_name;
    }

    inline const FrTOC::MapStatType_type& FrTOC::
    GetStatType( ) const
    {
      return m_StatType;
    }

#if 0
    class IFrameStream;
    class OFrameStream;

    class FrTOC: public FrBase
    {
    public:
      class TOCCallback;


      // order by data then nFrame
      typedef std::vector< position_list_type > position_dbl_list_type;

      struct ADC_type
      {
	INT_4U	channelID;
	INT_4U	groupID;
	position_list_type	positionADC;
      };

      typedef General::hash_map< std::string, ADC_type >
      MapADC_type;
      typedef std::vector< STRING >				IndexADC_type;

      static const CHAR* getStaticName();


      typedef General::hash_map< std::string, position_type >
      MapDetector_type;
      typedef std::vector< STRING >				IndexSer_type;
      typedef General::hash_map< std::string, position_list_type >
      MapSummary_type;

      struct Event_type
      {
	INT_4U		GTimeS;
	INT_4U		GTimeN;
	REAL_4		amplitude;
	position_type	position;

	inline Event_type( )
	{
	};

	inline Event_type( const General::GPSTime& GTime,
			REAL_4 	       Amplitude,
			position_type     Position )
	  : GTimeS( GTime.GetSeconds( ) ),
	    GTimeN( GTime.GetNanoseconds( ) ),
	    amplitude( Amplitude ),
	    position( Position )
	{
	}
      };

      typedef std::vector< Event_type >	EventList_type;

      typedef General::hash_map< std::string, EventList_type >
      MapEventType_type;
      typedef General::hash_map< std::string, EventList_type >
      MapSimEventType_type;

      FrTOC( );

      ~FrTOC( );

      std::string GetClassName( INT_2U ClassId ) const;

      INT_4U				GetNFrame( ) const;
      const std::vector< INT_4U >&	GetChannelID( ) const;
      const std::vector< INT_4U >&	GetFrame( ) const;
      const std::vector< INT_4U >&	GetGroupID( ) const;
      const std::vector< STRING >&	GetName( ) const;
      const position_list_type&		GetNFirstADC( ) const;
      const position_list_type&		GetNFirstMsg( ) const;
      const position_list_type&		GetNFirstSer( ) const;
      const position_list_type&		GetNFirstTable( ) const;
      INT_4U				GetNSH( ) const;
      INT_4U				GetNStatType( ) const;
      const position_dbl_list_type&	GetPositionADC( ) const;

      //:TODO: Put these into the correct place
      INT_4U				GetNADC( ) const;
      const MapADC_type&			GetADC( ) const;
      MapADC_type::const_iterator 		GetADC( const std::string& Name ) const;
      MapADC_type::const_iterator 		GetADC( INT_4U index ) const;
      INT_4U				GetNDetector( ) const;
      const MapDetector_type& 		GetDetector( ) const;
      INT_4U				GetNSummary( ) const;
      const MapSummary_type& 		GetSummary( ) const;
      
      INT_4U				GetNEventType( ) const;
      const MapEventType_type& 		GetEventType( ) const;
      MapEventType_type::const_iterator 	GetEventType( const std::string& Name ) const;

      INT_4U				GetNSimEventType( ) const;
      const MapSimEventType_type& 	GetSimEventType( ) const;
      MapSimEventType_type::const_iterator 	GetSimEventType( const std::string& Name ) const;
      //:TODO: End - Put these into the correct place

      void RegisterCallbacks( Dictionary& Dictionary, OFrameStream& Stream );

      void RegisterCallbacks( Dictionary& Dictionary, IFrameStream& Stream );

      void UnregisterCallbacks( );

      //: Read the structure from stream
      static FrTOC* ReadDynamic( IFrameStream& Stream,
				 const FrBase& Base );

      virtual void Write( OFrameStream& Stream ) const;

      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      friend class IFrameStream;
      friend class TOCCallback;

      position_list_type		m_positionH;
      position_list_type		m_nFirstADC;
      position_list_type		m_nFirstSer;
      position_list_type		m_nFirstTable;
      position_list_type		m_nFirstMsg;

      std::vector< INT_2U >	m_SHid;
      std::vector< STRING >	m_SHname;

      MapDetector_type		m_Detector;

      MapADC_type			m_ADC;
      IndexADC_type		m_ADCIndex;

      //:TODO: Put these into the correct place
      MapSummary_type		m_Summary;
      MapEventType_type		m_EventType;
      MapSimEventType_type	m_SimEventType;
      //:TODO: End - Put these into the correct place

      Dictionary*				m_dictionary;
      IFrameStream*				m_input_stream;
      OFrameStream*				m_output_stream;
      std::vector< Dictionary::callback_key_type >	m_callback_keys;
      Dictionary::Callback*			m_callback;

      void read( IFrameStream& Stream );

      void register_callbacks( );
    };

    inline bool FrTOC::StatTypeKey::
    operator!=( const StatTypeKey& RHS ) const
    {
      return ! ( *this == RHS );
    }

    inline INT_4U FrTOC::
    GetNADC( ) const
    {
      return m_ADC.size( );
    }

    inline const FrTOC::MapADC_type& FrTOC::
    GetADC( ) const
    {
      return m_ADC;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstADC( ) const
    {
      return m_nFirstADC;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstMsg( ) const
    {
      return m_nFirstMsg;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstSer( ) const
    {
      return m_nFirstSer;
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetNFirstTable( ) const
    {
      return m_nFirstTable;
    }

    inline INT_4U FrTOC::
    GetNFrame( ) const
    {
      return m_dataQuality.size( );
    }

    inline const FrTOC::position_list_type& FrTOC::
    GetPositionH( ) const
    {
      return m_positionH;
    }

    inline INT_4U FrTOC::
    GetNSH( ) const
    {
      return m_SHid.size( );
    }

    inline INT_4U FrTOC::
    GetNStatType( ) const
    {
      return m_StatType.size( );
    }

    inline const std::vector< INT_2U >& FrTOC::
    GetSHid( ) const
    {
      return m_SHid;
    }

    inline const std::vector< STRING >& FrTOC::
    GetSHname( ) const
    {
      return m_SHname;
    }

    //:TODO: Put these into there proper location
    inline INT_4U FrTOC::
    GetNDetector( ) const
    {
      return m_Detector.size( );
    }

    inline const FrTOC::MapDetector_type& FrTOC::
    GetDetector( ) const
    {
      return m_Detector;
    }

    inline INT_4U FrTOC::
    GetNSummary( ) const
    {
      return m_Summary.size( );
    }
    inline const FrTOC::MapSummary_type& FrTOC::
    GetSummary( ) const
    {
      return m_Summary;
    }

    inline INT_4U FrTOC::
    GetNEventType( ) const
    {
      return m_EventType.size( );
    }
    inline const FrTOC::MapEventType_type& FrTOC::
    GetEventType( ) const
    {
      return m_EventType;
    }

    inline INT_4U FrTOC::
    GetNSimEventType( ) const
    {
      return m_SimEventType.size( );
    }
    inline const FrTOC::MapSimEventType_type& FrTOC::
    GetSimEventType( ) const
    {
      return m_SimEventType;
    }

    //:TODO: End - Put these into there proper location
#endif /* 0 */

  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* FrameCPP_VERSION_6_FrTOC_HH */
