#ifndef FrameCPP_VERSION_6_FrTOCSimEvent_HH
#define FrameCPP_VERSION_6_FrTOCSimEvent_HH

#include "framecpp/Version6/STRING.hh"
#include "framecpp/Version6/GPSTime.hh"

namespace FrameCPP
{
  namespace Common
  {
    class TOCInfo;
  }

  namespace Version_6
  {
    //===================================================================
    //===================================================================
    class FrTOCSimEvent
    {
    public:
      typedef INT_4U	nevent_type;
      typedef STRING	name_type;
      typedef INT_4U	gtimesSim_type;
      typedef INT_4U	gtimenSim_type;
      typedef REAL_4	amplitudeSimEvent_type;
      typedef INT_8U	positionSimEvent_type;

      struct event_type {
	GPSTime			GTime;
	amplitudeSimEvent_type	amplitudeSimEvent;
	positionSimEvent_type	positionSimEvent;
      };

      static const int event_type_size
      = sizeof( gtimesSim_type )
	+ sizeof( gtimenSim_type )
	+ sizeof( amplitudeSimEvent_type )
	+ sizeof( positionSimEvent_type )
	;

      typedef std::vector< event_type > events_container_type;
      typedef std::map< name_type, events_container_type >
      nameSimEvent_container_type;

      FrTOCSimEvent( );

      Common::FrameSpec::size_type
      Bytes( const Common::StreamBase& Stream ) const;

      const nameSimEvent_container_type& GetSimEvent( ) const;

      void QuerySimEvent( const Common::TOCInfo& Info,
			  INT_4U FrameOffset,
			  INT_8U Position );

      bool operator==( const FrTOCSimEvent& RHS ) const;

      static void Description( Common::Description& Desc );

    protected:
      nameSimEvent_container_type		m_info;

      FrTOCSimEvent( Common::IStream& Stream );
      void write( Common::OStream& Stream ) const;
    };

    inline Common::FrameSpec::size_type FrTOCSimEvent::
    Bytes( const Common::StreamBase& Stream ) const
    {
      Common::FrameSpec::size_type 
	retval = sizeof( nevent_type );
      //-----------------------------------------------------------------
      // 
      //-----------------------------------------------------------------
      for ( nameSimEvent_container_type::const_iterator
	      cur = m_info.begin( ),
	      last = m_info.end( );
	    cur != last;
	    ++cur )
      {
	retval +=
	  cur->first.Bytes( )
	  + sizeof( nevent_type )
	  + ( cur->second.size( ) * event_type_size )
	  ;
      }
      return retval;
    }

    inline const FrTOCSimEvent::nameSimEvent_container_type& FrTOCSimEvent::
    GetSimEvent( ) const
    {
      return m_info;
    }

  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* FrameCPP_VERSION_6_FrTOCSimEvent_HH */
